const MaxNodeComm = require("../communication");
const MaxUtils = require("../utils");
const { EventEmitter } = require("events");

const { API_MESSAGES, ERROR_CODES, STATUS_CODES } = require("./constants.js");
const { NAMED_PIPE_PREFIX } = require("../constants");

const makeError = (message, code) => {
	const err = new Error(message);
	err.code = code;
	return err;
};

/**
 * @class APICommunicator
 */
class APICommunicator extends EventEmitter {

	constructor(options = {}) {

		super();

		this._options = Object.assign({
			maxConnections: 1,
			path: process.platform === "win32" ? `${NAMED_PIPE_PREFIX}${process.env.SCRIPT_ID}ctrl` : `/tmp/${process.env.SCRIPT_ID}_ctrl.sock`,
			protocol: MaxNodeComm.COMMUNICATION_TYPES.SOCKET,
			reqTimeout: 3000
		}, options);

		this._requestEvents = new EventEmitter();

		this._communicator = MaxNodeComm.createCommunicator(this._options.protocol, {
			path: this._options.path,
			port: this._options.port,
			maxConnections: this._options.maxConnections
		});

		this._communicator.on("client_connect", this._onClientConnect.bind(this));
		this._communicator.on("client_disconnect", this._onClientDisconnect.bind(this));
		this._communicator.on("client_message", this._onClientMessage.bind(this));

		this._communicator.on("close", this._onServerClose.bind(this));
		this._communicator.on("error", this._onServerError.bind(this));
	}

	/**
	 * @private
	 * @fires APICommunicator#connected
	 */
	_onClientConnect(client) {
		/**
		 * Commuicator got remote connection event
		 *
		 * @event APICommunicator#connected
		 */
		this.emit("connected");
	}

	/**
	 * @private
	 * @fires APICommunicator#disconnected
	 */
	_onClientDisconnect(client) {

		/**
		 * Commuicator lost remote connection event
		 *
		 * @event APICommunicator#disconnected
		 */
		this.emit("disconnected");
	}

	/**
	 * @private
	 * @param {Client} client The client sending the inlet message
	 * @param {String} uid The message uid
	 * @param {Object} payload The received payload
	 * @fires APICommunicator#inlet
	 */
	async _onInlet(client, uid, payload) {

		const inletValue = payload.value;

		if (!inletValue || (Array.isArray(inletValue) && inletValue.length === 0)) {
			try {
				await client.write({
					type: API_MESSAGES.INLET,
					uid,
					status: STATUS_CODES.ERROR,
					payload: {
						err: {
							message: "Invalid inlet message footprint: Missing value.",
							code: ERROR_CODES.INVALID_PARAMETERS,
							details: JSON.stringify(payload)
						}
					}
				});
			} catch (err) {
				// error while attempting to echo the inlet..Not really a big deal
			}
			return;
		}


		try {
			// echo the successful receiption
			await client.write({
				type: API_MESSAGES.INLET,
				uid,
				status: STATUS_CODES.SUCCESS,
				payload: {}
			});
		} catch (err) {
			// error while attempting to echo the inlet..Not really a big deal
		}

		/**
		 * Commuicator received inlet message event
		 *
		 * @event APICommunicator#inlet
		 */
		this.emit("inlet", inletValue);

	}

	/**
	 * @param {String} type The message type
	 * @param {Object} payload The payload to send
	 *
	 * @returns {Promise.<module:max-api.Response, Error>} Returns the response if resolved or an Error if rejected
	 */
	request(type, payload = {}) {
		return new Promise((resolve, reject) => {


			let respTimeout = null;
			const uid = MaxUtils.uniqID();


			const responseHandler = (status, responsePayload) => {
				clearTimeout(respTimeout);

				if (status === STATUS_CODES.ERROR) {
					const errDetails = responsePayload.error || {};
					reject(makeError(errDetails.message, errDetails.code));
					return;
				}

				if (status === STATUS_CODES.SUCCESS) {
					resolve(responsePayload);
					return;
				}

				// TODO: other STATUS_CODES!?
			};

			this._requestEvents.once(uid, responseHandler);

			this._communicator.broadcast({
				type,
				uid,
				payload
			})
				.then(() => {
					// nothing, we wait for the response
				})
				.catch((err) => {
					if (err) {
						this._requestEvents.removeAllListeners(uid, clearTimeout(respTimeout));
						reject(err);
						return;
					}
				});

			// Schedule Timeout
			setTimeout(() => {
				this._requestEvents.removeAllListeners(uid);
				reject(makeError(`Max API Message Request timed out. id: ${uid}`, ERROR_CODES.REQUEST_TIMEOUT));
				return;
			}, this._options.reqTimeout);

		});
	}

	/**
	 * Remote message handler
	 * @private
	 * @param {Object} msg The received msg object
	 * @param {Client} client The client sending the msg
	 */
	_onClientMessage(msg = {}, client) {
		const { uid, payload, status, type } = msg;

		switch (type) {

			// API Call Responses
			case API_MESSAGES.OUTLET: {
				this._requestEvents.emit(uid, status, payload);
				return;
			}

			case API_MESSAGES.POST: {
				this._requestEvents.emit(uid, status, payload);
				return;
			}

			case API_MESSAGES.GET_DICT: {
				this._requestEvents.emit(uid, status, payload);
				return;
			}

			case API_MESSAGES.SET_DICT: {
				this._requestEvents.emit(uid, status, payload);
				return;
			}

			// API Call Remote Requests
			case API_MESSAGES.INLET: {
				this._onInlet(client, uid, payload);
				return;
			}

			default:
				// TODO:
				console.log("Unknown message");
				console.log(msg);
				return;
		}
	}

	/**
	 * @private
	 * @fires APICommunicator#closed
	 */
	_onServerClose() {
		/**
		 * Communicator server closed Event
		 *
		 * @event APICommunicator#closed
		 */
		this.emit("closed");
	}

	/**
	 * @private
	 * @fires APICommunicator#error
	 */
	_onServerError(err) {
		/**
		 * Communicator server error Event
		 *
		 * @event APICommunicator#error
		 * @param {Error} err The occured error
		 */
		this.emit("error", err);
	}

	/**
	 * Close the communication channel.
	 * The server can be activated again using Bridge.listen
	 * @returns {Promise.<null, Error>} Returns null if resolved, or an Error if rejected.
	 */
	close() {
		this._requestEvents.removeAllListeners();
		return this._communicator.close();
	}

	/**
	 * Destroy the communication channel entirely.
	 * The server cannot be reactivated after this.
	 * @returns {Promise.<null, Error>} Returns null if resolved, or an Error if rejected.
	 */
	destroy() {
		this._requestEvents.removeAllListeners();
		return this._communicator.destroy();
	}

	/**
	 * Activate/Start the communication channel with Max
	 * @returns {Promise.<Object, Error>} Returns the domain specific connections details if resolved, or an Error if rejected.
	 */
	listen() {
		return this._communicator.listen();
	}

}

module.exports = new APICommunicator();
